/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
    Contributing author: Nicolas Onofrio, Purdue University
       (nonofrio@purdue.edu)
------------------------------------------------------------------------- */

#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "compute_laplacian.h"
#include "atom.h"
#include "comm.h"
#include "pair.h"
#include "force.h"
#include "error.h"
#include "memory.h"
#include "modify.h"
#include "neighbor.h"
#include "neigh_list.h"
#include "neigh_request.h"
#include "math.h"
#include "group.h"
#include "compute.h"
#include "update.h"
#include <iostream>
#include <fstream>

using namespace LAMMPS_NS;

/* ---------------------------------------------------------------------- */

ComputeLap::ComputeLap(LAMMPS *lmp, int narg, char **arg) :
  Compute(lmp, narg, arg)
{
  if (narg > 10) error->all(FLERR,"Illegal compute laplacian command");

  rc = force->numeric(FLERR,arg[3]);
  norm = force->numeric(FLERR,arg[4]);
  
  flagbound = 0;
  if (narg > 5) {
    if (strcmp(arg[5], "boundary") == 0) {  
      flagbound = 1;
      groupbit1 = group->bitmask[group->find(arg[6])];
      groupbit2 = group->bitmask[group->find(arg[8])];
      loc01 = force->numeric(FLERR,arg[7]);
      loc02 = force->numeric(FLERR,arg[9]);
    }
  }

// Delete current file 
  FILE *f;
  using namespace std;
  ifstream fin("current");
  if (fin){
      remove("current");
      fin.close();
  }

  comm_forward = 3;

}

/* ---------------------------------------------------------------------- */

ComputeLap::~ComputeLap()
{
}

/* ---------------------------------------------------------------------- */

void ComputeLap::init()
{
  int irequest = neighbor->request(this);
  neighbor->requests[irequest]->pair = 0;
  neighbor->requests[irequest]->compute = 1;
  neighbor->requests[irequest]->half = 1;
  neighbor->requests[irequest]->full = 0;
}

/* ---------------------------------------------------------------------- */

void ComputeLap::setup()
{
  compute_peratom();
}

/* ---------------------------------------------------------------------- */

void ComputeLap::compute_peratom()
{
  int inum,jnum,i,j,ii,jj;
  int *ilist,*jlist,*numneigh,**firstneigh;
  double xtmp,ytmp,ztmp,delx,dely,delz,rsq,w,n,we,ce;
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;
  double *loc, *lap;

  FILE *f;
  f = fopen("current", "a");

  int size, rank;
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);

// Get loc from fix/property
  get_names("loc",loc); 
  get_names("lap",lap); 

// Call neighbor list
  Pair *epair = force->pair;
  inum = epair->list->inum;
  ilist = epair->list->ilist;
  numneigh = epair->list->numneigh;
  firstneigh = epair->list->firstneigh;

// Set boundary conditions if flagbound = 1 
  double cumul1 = 0.0;
  double cumul2 = 0.0;
  double dt = update -> dt;
  if (flagbound == 1) {
    for (i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit1) {
        cumul1 += (loc01-loc[i])/dt;
        loc[i] = loc01;
      }
      if (mask[i] & groupbit2) {
        cumul2 += (loc02-loc[i])/dt;
        loc[i] = loc02;
      }
    }
  }

// Forward loc to ghost
  comm->forward_comm_compute(this);

// Set laplacian and norm to 0
  for (i = 0; i < nlocal; i++) {
    lap[i]=0.0;
  }

  n = 0.0;
// Loop over local i
  for (ii = 0; ii < inum; ii++) {
    i = ilist[ii];
    if (mask[i] & groupbit) {

      jlist = firstneigh[i];
      jnum = numneigh[i]; 

      xtmp = x[i][0];
      ytmp = x[i][1];
      ztmp = x[i][2];

// Loop over neighbor j of i
      for (jj = 0; jj < jnum; jj++) {
        j = jlist[jj];
        j &= NEIGHMASK;
        if (mask[j] & groupbit) {

          delx = xtmp - x[j][0];
          dely = ytmp - x[j][1];
          delz = ztmp - x[j][2];
          rsq = delx * delx + dely * dely + delz * delz;

// Weighting function and normalization
          if (rsq < rc*rc) w = (1-(rsq)/(rc*rc))*(1-(rsq)/(rc*rc));
          else w = 0.0;

          n += w;
          if (j < nlocal) n += w;

// Compute laplacian for i and j because half list
          lap[i] += w * (loc[j] - loc[i]) / rsq;
          if (j < nlocal) lap[j] += w * (loc[i] - loc[j]) / rsq;

        } 
      }
    } 
  }

// Compute normalization
//  n = n / (6.0 * nlocal);
//  if (flagnorm == 1) norm = n;
//  printf("norm  %f",n);
  for (i = 0; i < nlocal; i++) {
    lap[i] = lap[i] / norm;
  }

// Write to the current file
//  printf("cumul1 %f", cumul1);
  double local[2] = {0,0};
  double global[2] = {0,0};
  local[0] = cumul1;
  local[1] = cumul2;
  MPI_Reduce(local,global,2,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  if (rank == 0 && update->ntimestep % 10 == 0) fprintf(f,"Step %lu Electrodes ce: %f we: %f total: %f\n",update->ntimestep,global[0],global[1],global[0]+global[1]);
  fclose(f);

}

/* ---------------------------------------------------------------------- */

void ComputeLap::get_names(char *c,double *&ptr)
{
 int index,flag;
 index = atom->find_custom(c,flag);

 if(index!=-1) ptr = atom->dvector[index];
 else error->all(FLERR,"Compute laplacian requires fix property/atom d_loc d_lap command");
}

/* ---------------------------------------------------------------------- */

int ComputeLap::pack_forward_comm(int n, int *list, double *buf, int pbc_flag, int *pbc)
{
  int i,j,m;
  double *loc;
  get_names("loc",loc);

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    buf[m++] = loc[j];
  }
  return m;
}

/* ---------------------------------------------------------------------- */

void ComputeLap::unpack_forward_comm(int n, int first, double *buf)
{
  int i,m,last;
  double *loc;
  get_names("loc",loc);

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) loc[i] = buf[m++];
}

/* ---------------------------------------------------------------------- */
