/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Nicolas Onofrio, Purdue University
   (nonofrio@purdue.edu)
------------------------------------------------------------------------- */

#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "fix_echemdid.h"
#include "force.h"
#include "error.h"
#include "memory.h"
#include "modify.h"
#include "math.h"
#include "group.h"
#include "atom.h"
#include "compute.h"
#include "update.h"

using namespace LAMMPS_NS;
using namespace FixConst;

/* ---------------------------------------------------------------------- */

FixEChemDID::FixEChemDID(LAMMPS *lmp, int narg, char **arg) :
  Fix(lmp, narg, arg)
{
  if (narg < 12) error->all(FLERR,"Illegal fix EChemDID command");

// General argument for diffusion of loc
// fix ID groupID EChemDID nevery k k_value cut cut_value norm v_norm [boundary g1 v1 g2 v2]
  nevery = atoi(arg[3]);
  if (nevery <= 0) error->all(FLERR,"Illegal fix EChemDID command");

  if (force->numeric(FLERR,arg[5]) <= 0) error->all(FLERR,"Illegal fix EChemDID command");
  else k = force->numeric(FLERR,arg[5]);

  if (force->numeric(FLERR,arg[7]) <= 0) error->all(FLERR,"Illegal fix EChemDID command");

  if (force->numeric(FLERR,arg[9]) <= 0) error->all(FLERR,"Illegal fix EChemDID command");

  if (force->numeric(FLERR,arg[11]) <= 0) error->all(FLERR,"Illegal fix EChemDID command");
  else nelec = force->numeric(FLERR,arg[11]);

// Flag for compute laplacian
  flagbound = 0;
  flagdamp = 0;
  loc0 = 0.0;
  eta = 0.0;

// If boundary constrain
  if (narg > 12) {
    if (strcmp(arg[12], "boundary") == 0) flagbound = 1;
    else error->all(FLERR,"Illegal fix EChemDID command option"); 
  }

  if (flagbound != 0) {
    if (group->find(arg[13]) == -1) error->all(FLERR,"Could not find EChemDID boundary group ID");

    if (group->find(arg[15]) == -1) error->all(FLERR,"Could not find EChemDID boundary group ID");
  }

// Damped EChemDID
// If arg "eta" then damped EChemDID
  if (narg > 17) {
      flagdamp = 1;
      loc0 = (force->numeric(FLERR,arg[14])+force->numeric(FLERR,arg[16]))/2.;
      eta = force->numeric(FLERR,arg[18]);
  }
  rc = force->numeric(FLERR,arg[7]);

// Create a compute ID = fixID_lap
  int n = strlen(id) + 5;
  id_lap = new char[n];
  strcpy(id_lap,id);
  strcat(id_lap,"_lap");

// Call the compute constrain or not on boundary
  if (flagbound == 1) {
    char **newarg = new char*[10];
    newarg[0] = id_lap;
    newarg[1] = arg[1]; // groupbit
    newarg[2] = (char *) "laplacian";
    newarg[3] = arg[7]; // rc
    newarg[4] = arg[9]; // norm
    newarg[5] = arg[12]; // boundary
    newarg[6] = arg[13]; // group bound1
    newarg[7] = arg[14]; // loc01 init
    newarg[8] = arg[15]; // group bound2
    newarg[9] = arg[16]; // loc02 init
    modify->add_compute(10,newarg);
    delete [] newarg;
  } else {
    char **newarg = new char*[5];
    newarg[0] = id_lap;
    newarg[1] = arg[1];
    newarg[2] = (char *) "laplacian";
    newarg[3] = arg[7];
    newarg[4] = arg[9];
    modify->add_compute(5,newarg);
    delete [] newarg;
  }

}

/* ---------------------------------------------------------------------- */

FixEChemDID::~FixEChemDID()
{
  modify->delete_compute(id_lap);
  delete [] id_lap;
}

/* ---------------------------------------------------------------------- */

int FixEChemDID::setmask()
{
  int mask = 0;
  mask |= POST_INTEGRATE;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixEChemDID::init()
{
  int icompute = modify->find_compute(id_lap);
  if (icompute < 0)
    error->all(FLERR,"Laplacian ID for fix EChemDID does not exist");
  laplacian = modify->compute[icompute];
}

/* ---------------------------------------------------------------------- */

void FixEChemDID::post_integrate()
{
  int i, j;
  int nlocal = atom -> nlocal;
  double *loc, *lap, *Wi;
  double dt = update -> dt;
  double W0, Fcut;
  int *mask = atom->mask;

  W0 = (1.-(rc-rc*0.01)/(rc*rc))*(1-(rc-rc*0.01)/(rc*rc));
  get_names("loc",loc);
  get_names("lap",lap);
  get_names("Wi",Wi);

  for (i = 0; i < nelec; i++) {
    laplacian -> compute_peratom();
    for (j = 0; j < nlocal; j++) {
      if (mask[j] & groupbit) {
        if (flagdamp == 1) {
          if (Wi[j] < W0) Fcut = (1.-(Wi[j]/W0)*(Wi[j]/W0))*(1.-(Wi[j]/W0)*(Wi[j]/W0));
          else Fcut = 0.0;
        } else Fcut = 0.0;
        loc[j] += k*dt*lap[j]/nelec - eta*(loc[j]-loc0)*Fcut*dt/nelec;
      }
    }
  }

  
}

/* ---------------------------------------------------------------------- */

void FixEChemDID::get_names(char *c,double *&ptr)
{
 int index,flag;
 index = atom->find_custom(c,flag);

 if(index!=-1) ptr = atom->dvector[index];
 else error->all(FLERR,"Compute laplacian requires fix property/atom d_loc d_lap d_Wi command");
}

/* ---------------------------------------------------------------------- */
